#ifndef SCENE_H
#define SCENE_H

#include <glm/mat4x4.hpp>
#include <vector>
#include <unordered_map>
#include "light.h"

class Camera;
class Pipeline;
class Mesh;

// Scene interface :

template <class T>
class SceneIterator
{
public:
    virtual SceneIterator& operator++() = 0;
    virtual T operator*() = 0;
    virtual bool isValid() = 0;
    void next() {if(isValid()) operator++();}
    T getItem() {return operator*();}
};

struct GeometryNode
{
    Mesh* mesh;
    glm::mat4 modelMatrix;
    
    GeometryNode(Mesh* m, glm::mat4 transform) : mesh(m), modelMatrix(transform) {}
};

class Scene
{
protected:
    Pipeline *m_pipeline;

public:
    Scene() : m_pipeline(NULL) {}

    Pipeline* getPipeline() {return m_pipeline;}

    virtual SceneIterator<Light*> getLights(unsigned int flags) = 0;
    virtual SceneIterator<GeometryNode*> getGeometry(unsigned int flags) = 0;

    virtual void getMeshTypes(std::vector<unsigned int> &meshTypes);
    virtual void getLightTypes(std::vector<unsigned int> &lightTypes);
};

// A basic implementation :

template <class T>
class ArrayIterator : public SceneIterator<T>
{
    std::vector<T> &vec;
    unsigned int id;
public:
    ArrayIterator(std::vector<T> &myVec, int myId=0) : vec(myVec), id(myId) {}
    virtual SceneIterator<T>& operator++() {++id; return *this;}
    virtual T operator*() {return vec[id];}
    virtual bool isValid() {return id < vec.size();}
};

class ArrayScene : public Scene
{
protected:
    union Node
    {
        Light* light;
        GeometryNode* geometry;
    };

    std::unordered_map<std::vector<Node>*> nodes;
    
    std::vector<GeometryNode*> allocations;
public:
    ArrayScene() : Scene() {}
    ~ArrayScene();

    void setPipeline(Pipeline *pipeline) {m_pipeline = pipeline;}
    void clearScene();
    void addMesh(GeometryNode* node);
    void addMesh(Mesh* m, glm::mat4 transform);
    void addLight(Light* myLight) {lights.push_back(myLight);}
    Mesh* getMesh(int id) {return geometry[id]->mesh;}

    SceneIterator<Light*> getLights(unsigned int flags);
    SceneIterator<GeometryNode*> getGeometry(unsigned int flags);
};

#endif // SCENE_H
