#include <glm/mat4x4.hpp>
#include <glm/mat3x3.hpp>
#include <glm/ext.hpp>
#include "skyboxmodule.h"
#include "shader.h"
#include "texture.h"
#include "camera.h"
#include "framebuffer.h"

#include "sparrowrenderer.h"

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

SkyboxModule::SkyboxModule(Texture* myCubeMap)
{
    cubeMap = myCubeMap;

    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);
    glGenBuffers(2, vbos);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbos[0]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 36 * sizeof(GLubyte), skyboxIndices, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, vbos[1]);
    glBufferData(GL_ARRAY_BUFFER, 24 * sizeof(GLfloat), skyboxVertices, GL_STATIC_DRAW);

    if(SparrowRenderer::isModernOpenGLAvailable())
    {
        renderTarget = FrameBuffer::screen;
        shader = new Shader(vertSource, fragSource);
        mvpLocation = shader->getLocation("MVP");
    }

    glBindVertexArray(0);
}

SkyboxModule::~SkyboxModule()
{
    if(SparrowRenderer::isModernOpenGLAvailable())
    {
        glDeleteVertexArrays(1, &vao);
        glDeleteBuffers(2, vbos);
    }
}

void SkyboxModule::renderGL(Camera* myCamera, Scene* scene)
{
    glViewport(0, 0, width, height);
    glm::mat4 viewMatrix = glm::mat4(glm::mat3(myCamera->getViewMatrix()));
    glm::mat4 projectionMatrix = myCamera->getProjectionMatrix();
    glDisable(GL_CULL_FACE);
    glDepthMask(GL_FALSE);
    if(!SparrowRenderer::isModernOpenGLAvailable())
    {
        glMatrixMode(GL_MODELVIEW);
        glLoadMatrixf(glm::value_ptr(viewMatrix));
        glMatrixMode(GL_PROJECTION);
        glLoadMatrixf(glm::value_ptr(projectionMatrix));
    }
    else
    {
        renderTarget->bindFBO();
        shader->bind();
        shader->bindMat4(mvpLocation, projectionMatrix * viewMatrix);
    }
    cubeMap->bind(0);
    glBindVertexArray(vao);

    glBindBuffer(GL_ARRAY_BUFFER, vbos[1]);
    if(SparrowRenderer::isModernOpenGLAvailable())
    {
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3*sizeof(GLfloat), BUFFER_OFFSET(0));
        glEnableVertexAttribArray(0);
    }
    else
    {
        glDisable(GL_LIGHTING);
        glEnable(GL_TEXTURE_CUBE_MAP);

        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(3, GL_FLOAT, 0, BUFFER_OFFSET(0));
    }

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbos[0]);
    glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_BYTE, NULL);
    glBindVertexArray(0);

    if(!SparrowRenderer::isModernOpenGLAvailable())
    {
        glEnable(GL_LIGHTING);
        glDisable(GL_TEXTURE_CUBE_MAP);
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    glEnable(GL_CULL_FACE);
    glDepthMask(GL_TRUE);
}

void SkyboxModule::setRenderTarget(const FrameBuffer* target)
{
    if(target != NULL)
        renderTarget = target;
}

const GLfloat SkyboxModule::skyboxVertices[] = {
    -1.0f,  1.0f,  1.0f,
    -1.0f,  1.0f, -1.0f,
    -1.0f, -1.0f,  1.0f,
    -1.0f, -1.0f, -1.0f,
     1.0f,  1.0f,  1.0f,
     1.0f,  1.0f, -1.0f,
     1.0f, -1.0f,  1.0f,
     1.0f, -1.0f, -1.0f
};

const GLubyte SkyboxModule::skyboxIndices[] = {
    0, 6, 2,
    0, 4, 6,
    1, 3, 5,
    3, 7, 5,
    0, 3, 1,
    0, 2, 3,
    4, 5, 6,
    5, 7, 6,
    0, 5, 4,
    0, 1, 5,
    2, 6, 3,
    6, 7, 3
};

const std::string SkyboxModule::vertSource =
        "layout(location = 0)in vec3 inPosition;\n\
        out vec3 varTexCoord;\n\
        uniform mat4 MVP;\n\
        void main()\n\
        {\n\
            gl_Position = MVP * vec4(inPosition, 1.0);\n\
            varTexCoord = inPosition;\n\
        }\n";

const std::string SkyboxModule::fragSource =
        "in vec3 varTexCoord;\n\
        out vec4 outColor;\n\
        uniform samplerCube skybox;\n\
        void main()\n\
        {\n\
            outColor = texture(skybox, varTexCoord);\n\
        }\n";
